//------------------------------------------------------------------------
//
// Booster_BiDi (neu)
//
// Copyright (c) 2009 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      Booster_BiDi.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2007-08-29 V0.01 start
//            2008-10-01 V0.02 modified for Booster2, no cutout yet.
//            2009-03-17 V0.03 added immediate try to ENABLE if dcc_state
//                             changes from 0 to 1.
//            2009-04-14 V0.04 dcc interrupt added
//            2009-05-11 V0.05 Debug LEDs added, shorter int-dead time
//            2009-05-13 V0.06 rewritten: DCC changed to timing gap evaluation
//                             cutout done with interrupts
//                             added PLL code to get in sync with incoming DCC
//            2009-05-17 V0.07 digital filter instead of PLL
//                             yellow LED now BiDi Rec
//            2009-06-21 V1.20 block_short neu hinzu.
//            2009-07-24 V1.21 FEEDBACK neu dazu - SJ3 and SJ4 are open!!!
//            2009-09-14 V1.22 externes Ein/Ausschalten ber PD6 (EXTON)
//            2010-02-09 V1.23 bugfix: cutout was without power ENABLED
//            2010-03-17 V1.24 minor bugfix: longer LED_RED
//
// tests:     2009-05-17 BiDi created
//            2009-05-21 BiDi received
//            
//------------------------------------------------------------------------
//
// purpose:   A: Intelligent Timer for turning off the Booster
//            B: Create Cutout for BiDi (Railcom (R))
//
// target:    AVR Attiny2313, running @8MHz internal clock.
//          
// content:    1. Defines and variable definitions
//             2. Init
//             3  PLL: adjust OSCCAL to DCC Input
//             4. ISR (only timerval++ and DCC)
//             5. Tools (timeout, blinking)
//             6. MAIN: check the states, trace and retry enable
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>

//==============================================================================
//
// Section 1   Defines
//
//==============================================================================

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine (toggle fast)
                                 // 3: test motor

#define DCC_RAILCOM  1           // 1: receive of railcom supported

//---------------------------------------------------------------------
// Timing Definitions: (all values in us)
//
#ifndef F_CPU
   // prevent compiler error by supplying a default 
   # warning "F_CPU not defined for <booster_timer.c>", set default to 8MHz 
   # define F_CPU 8000000UL
   // if changed: check every place where it is used
   // (possible range underflow/overflow in preprocessor!)
#endif
#if (F_CPU > 30000000)
  #warning possible long int overflow - please check all define calculations
#endif
//---------------------------------------------------------------------
//
// Controlling the H-Bridge:
//
// a) DCC failure:
//    DCC must have a change every 12ms; if there is no change for 15ms,
//    turn off; turn on after a change (DCC state)
//
// b) Short detection:
//    after a short is detected for more than 100us (this time is done via HW)
//    turn off H-bridge.
//    try to turn on again: three times after 4ms
//    if this fails: wait 1s, then try again.
//
//---------------------------------------------------------------------
//
// Connect the Attiny2313 as follows:
// (outer pin mumbers: old 74123 Pinning)
//
//                        --------U--------
//                       -| 1 PA2   VCC 20|- ------------->>  
//   SHORT(in, low)      -| 2 PD0   PB7 19|- 
//   LOADED(in, low)     -| 3 PD1   PB6 18|-  
//                       -| 4 PA1   PB5 17|-  
//   LED orange(out)     -| 5 PA0   PB4 16|-
//   DCC(in, INT0)       -| 6 PD2   PB3 15|-  LED grn (active low)
//   FEEDBACK (out,2.01) -| 7 PD3   PB2 14|-  JUMPER (in)
//   LED rot(out, low)   -| 8 PD4   PB1 13|-  
//   ENABLE (out, hi)    -| 9 PD5   PB0 12|-  DATA_ON (out) (low=cutout)
//            |--------- -|10 GND   PD6 11|-  EXTON (in, high = on)
//                        -----------------
//
//
// This design must run with internal clock (XTAL1 is used as IO)
//
// JUMPER: if fitted: create CUTOUT
//
//---------------------------------------------------------------------
// Definition for DCC-Timing


/// This are DCC timing definitions from NMRA
#define PERIOD_1   116L          // 116us for DCC 1 pulse - do not change
#define PERIOD_0   232L          // 232us for DCC 0 pulse - do not change
#define MAX_DCC_SIZE  6

#define RAILCOM_CUTOUT_START   (30L-13L)  // min. 30us from Endbit until Datagram1
                                          // sub. 13: latency from ISR
#define RAILCOM_CUTOUT_LENGTH  (4*PERIOD_1 - RAILCOM_CUTOUT_START - 16) // 4 Bits, stop 16us before next slope
                                                                          // 6us nominal too short
#define RAILCOM_DEADTIME    4                                            // 4us take over

//---------------------------------------------------------------------
// Definition for OUTPUT and Timing

#define TICK_PERIOD              1000L       // 1ms tick for Timer

#define DCC_TIMEOUT_PERIOD      15000L       // 15ms

#define FAST_RECOVER_PERIOD      4000L

#define SLOW_RECOVER_PERIOD   1000000L 

#define LED_RED_HOLD_TIME      400000L 

#define BLINK_PERIOD           600000L

#define BLOCK_SHORT_TIME         8000L

//---------------------------------------------------------------------------
// PORT Definitions:
//
// PORTB:
//
#define DATA_ON     0     // out, if low: create cutout  
#define NU_PB1      1     // in, unused
#define JUMPER      2     // in, internal pullup
#define LED_GREEN   3     // out, active low
#define NU_PB4      4     // out, unused
#define NU_PB5      5     // out, unused
#define NU_PB6      6     // out, unused
#define NU_PB7      7     // out, unused

// PORTD:

#define SHORT       0     // in: active low
#define LOADED      1     // in: active low
#define DCC_IN      2     // in, pullup
#define FEEDBACK    3     // out, active low
#define LED_RED     4     // out, active low
#define ENABLE      5     // out: active high
#define EXTON       6     // in: pullup, low = on, high = off
//#define NU_PD6      6     // out, unused (1ms Tick)

// PORTA:

#define LED_ORANGE  0     // out, active low


#define DCC_IN_HIGH        ( (PIND & (1<<DCC_IN)))
#define DCC_IN_LOW         (!(PIND & (1<<DCC_IN)))
#define DCCIN_STATE        (  PIND & (1<<DCC_IN) )

#define LED_GREEN_OFF       PORTB |= (1<<LED_GREEN)
#define LED_GREEN_ON        PORTB &= ~(1<<LED_GREEN)

#define LED_ORANGE_OFF      PORTA |= (1<<LED_ORANGE)
#define LED_ORANGE_ON       PORTA &= ~(1<<LED_ORANGE)

#define LED_RED_OFF         PORTD |= (1<<LED_RED)
#define LED_RED_ON          PORTD &= ~(1<<LED_RED)
#define IS_LED_RED_ON()     (PORTD & (1<<LED_RED))


#define IS_SHORT            (!(PIND & (1<<SHORT)))
#define IS_LOADED           (!(PIND & (1<<LOADED)))

#define JUMPER_FITTED       (!(PINB & (1<<JUMPER)))

#define IS_EXTON            ((PIND & (1<<EXTON)))

#define NO_FEEDBACK         PORTD |= (1<<FEEDBACK)
#define MAKE_FEEDBACK       PORTD &= ~(1<<FEEDBACK)


#define ENABLE_OUT          PORTD |= (1<<ENABLE)
#define DISABLE_OUT         PORTD &= ~(1<<ENABLE)
#define ENABLED             (PORTD & (1<<ENABLE))        // as state var

#define NO_CUTOUT           PORTB |= (1<<DATA_ON)
#define MAKE_CUTOUT         PORTB &= ~(1<<DATA_ON)
#define IS_CUTOUT           (!(PORTB & (1<<DATA_ON)))    // as state var

//----------------------------------------------------------------------------
// DEBUG LEDs
// (Note: if a debug LED is not need, simply make an empty #define)
// LED_DCC_POLA   = show input polarity
// LED_BIDI_REC   = show BiDi received
// LED_RAMP_**    = behavior of PLL for DCC Tracking
// LED_DCC_ST_BIT = DCC Start Bit
// LED_TIMER_TICK = Timer Tick   ____|_____|_____|___ (1ms)
// LED_DCC_EDGE   = DCC Edge

//#define LED_DCC_POLA_P      PORTB |= (1<<NU_PB7)            // Pin 19
//#define LED_DCC_POLA_N      PORTB &= ~(1<<NU_PB7)
#define LED_DCC_POLA_P
#define LED_DCC_POLA_N

#define LED_PLL_CALL          PORTB |= (1<<NU_PB7)            // Pin 19
#define LED_PLL_CALL_OFF      PORTB &= ~(1<<NU_PB7)


#define LED_BIDI_REC       PORTB |= (1<<NU_PB6)           // Pin 18
#define LED_BIDI_REC_OFF   PORTB &= ~(1<<NU_PB6)
//#define LED_BIDI_REC       
//#define LED_BIDI_REC_OFF   

#define LED_RAMP_UP        {PORTB |=  (1<<NU_PB5); PORTB &= ~(1<<NU_PB4); }
#define LED_RAMP_DOWN      {PORTB &= ~(1<<NU_PB5); PORTB |=  (1<<NU_PB4); }
#define LED_RAMP_OFF       {PORTB &= ~(1<<NU_PB5); PORTB &= ~(1<<NU_PB4); }

//#define LED_DCC_EDGE        PORTB |= (1<<NU_PB5)            // Pin 17
//#define LED_DCC_EDGE_OFF    PORTB &= ~(1<<NU_PB5)
#define LED_DCC_EDGE
#define LED_DCC_EDGE_OFF

#define LED_DCC_ON          PORTB |= (1<<NU_PB1)            // Pin 13
#define LED_DCC_OFF         PORTB &= ~(1<<NU_PB1)

//#define LED_DCC_ST_BIT      PORTB |= (1<<NU_PB6)            // Pin 18
//#define LED_DCC_ST_BIT_OFF  PORTB &= ~(1<<NU_PB6)
#define LED_DCC_ST_BIT
#define LED_DCC_ST_BIT_OFF


//#define LED_DCC_END_BIT     PORTB |= (1<<NU_PB4)            // Pin 16
//#define LED_DCC_END_BIT_OFF PORTB &= ~(1<<NU_PB4)
#define LED_DCC_END_BIT
#define LED_DCC_END_BIT_OFF

//#define LED_TIMER_TICK      PORTB |= (1<<NU_PB4)            // Pin 16
//#define LED_TIMER_TICK_OFF  PORTB &= ~(1<<NU_PB4)
#define LED_TIMER_TICK 
#define LED_TIMER_TICK_OFF 


//----------------------------------------------------------------------------
// Internal Hardware usage:
//
// Timer 1: prescaler 8, running @ 1MHz, normal mode. 
//          permanently running, wrap around; 
//          OCRx used to create cutout
//
//----------------------------------------------------------------------------
// Global Data

#define OPTIMIZE_VARS 1

#if (OPTIMIZE_VARS == 1)
     #define Communicate GPIOR0
     #define Recstate GPIOR1
     #define DCCstate GPIOR2
#else
     unsigned char Communicate;
     unsigned char Recstate; 
     unsigned char DCCstate;
#endif

// timer access
typedef union { unsigned int ui16;
                signed int i16;
                struct {  signed int i10: 10;
                          unsigned char rest: 4;
                       };
                struct {  signed char  i8l;
                          signed char  i8h;
                       };
                struct {  unsigned char  ui8l;
                          unsigned char  ui8h;
                       };
              } t_timer;

t_timer dcc_edge;         // loaded by ISR
t_timer last_dcc_edge;
t_timer previous_dcc_edge;
t_timer pre_previous_dcc_edge;

signed int period_1_4x = PERIOD_1 * 4;
signed int period_1_1x = PERIOD_1;


// t_timer railcom_refpoint;      // loaded by DCC Tracker


unsigned char version[] PROGMEM    = {" -- Booster_V2.01 V1.24 (BiDi)"};
unsigned char version1[] PROGMEM    = {" -- block10ms, PD6, CUTOUT "};
unsigned char copyright1[] PROGMEM = {" -- (c) Kufer Wolfgang, 17.03.2010 "};
unsigned char copyright2[] PROGMEM = {" -- www.opendcc.de -- "};

//----------------------------------------------------------------------------------
// Definitions for inter process communication


#define C_Received        0    // a new DCC was received - issued by ISR(Timer1)
                               //                          cleared by main
#define C_Railcom         1    // Railcom window is open
                               //                        - issued by ISR
                               //                          cleared by watch_railcom
#define C_RailcomRec      2    // we are receiving a cutout

#define C_DCC_Edge        3    // a new DCC edge was received - issued by ISR(INT0)
                               //                          cleared by clock adjust

#define C_Out_OFF         4    // true, if output is currently off - issued by tracker
                               //                          used by cutout




static inline unsigned char semaphor_query(unsigned char flag) 
       __attribute__((always_inline));

unsigned char semaphor_query(unsigned char flag)
  {
    return (Communicate & (1<<flag));
  }
               
static inline void semaphor_set(unsigned char flag) 
       __attribute__((always_inline));

void semaphor_set(unsigned char flag)
  {
    cli();
    Communicate |= (1<<flag);
    sei();
  }

static inline void semaphor_set_isr(unsigned char flag) 
       __attribute__((always_inline));

void semaphor_set_isr(unsigned char flag)
  {
    Communicate |= (1<<flag);
  }
       
static inline unsigned char semaphor_get(unsigned char flag) 
       __attribute__((always_inline));

unsigned char semaphor_get(unsigned char flag)
  {
    unsigned char value;
    cli();
    value = Communicate & (1<<flag);
    Communicate &= (~(1<<flag));
    sei();
    return(value);
  }

static inline void semaphor_clear(unsigned char flag) 
       __attribute__((always_inline));

void semaphor_clear(unsigned char flag)
  {
    cli();
    Communicate &= (~(1<<flag));
    sei();
  }

//==============================================================================
//
// Section 2   INIT
//
//==============================================================================


void init_main(void)
  {
    DDRB  = (1 << DATA_ON )         // out, if low: create cutout  
          | (1 << NU_PB1  )         // out, unused
          | (0 << JUMPER   )        // in, internal pullup
          | (1 << LED_GREEN  )      // out, active low
          | (1 << NU_PB4 )          // out, unused
          | (1 << NU_PB5  )         // out, unused
          | (1 << NU_PB6  )         // out, unused
          | (1 << NU_PB7   );       // out, unused


    PORTB  = (1 << DATA_ON )        // out, if low: create cutout  
          | (0 << NU_PB1  )         // out, unused
          | (1 << JUMPER   )        // in, internal pullup
          | (1 << LED_GREEN  )      // out, active low
          | (0 << NU_PB4 )          // out, unused
          | (0 << NU_PB5  )         // out, unused
          | (0 << NU_PB6  )         // out, unused
          | (0 << NU_PB7   );       // out, unused
          

    DDRD  = (0<<  SHORT )           // in: active low
        | (0<<  LOADED  )           // in: active low
        | (0<<  DCC_IN  )           // in, pullup
        | (1<<  FEEDBACK )          // out, active low
        | (1<<  LED_RED )           // out, active low
        | (1<<  ENABLE  )           // out: active high
//        | (1<<  NU_PD6  );        // out, unused
        | (0<<  EXTON  );           // in
    
    PORTD  = (1<<  SHORT )          // in: active low -> pullup
        | (1<<  LOADED  )           // in: active low
        | (1<<  DCC_IN  )           // in, pullup
        | (1<<  FEEDBACK  )         // out, active low (not activated)
        | (0<<  LED_RED )           // out, active low
        | (0<<  ENABLE  )           // out: active high
//        | (0<<  NU_PD6  );        // out, unused
        | (1<<  EXTON );            // in, pullup (default: on)
    

    DDRA  = (1 << LED_ORANGE);        //    
    
    PORTA = (0 << LED_ORANGE);        

    // LEDs all on


    // Init Timer0

    #define T0_PRESCALER   8    // may be 1, 8, 64, 256, 1024
    #if   (T0_PRESCALER==1)
        #define T0_PRESCALER_BITS   ((0<<CS02)|(0<<CS01)|(1<<CS00))
    #elif (T0_PRESCALER==8)
        #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(0<<CS00))
    #elif (T0_PRESCALER==64)
        #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(1<<CS00))
    #elif (T0_PRESCALER==256)
        #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(0<<CS00))
    #elif (T0_PRESCALER==1024)
        #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(1<<CS00))
    #endif

//    #define T87US (F_CPU * PERIOD_1 * 3 / 4 / T0_PRESCALER / 1000000L)

    #define T87US (F_CPU * 77L / T0_PRESCALER / 1000000L)

    #if (T87US > 254)
      #warning T87US too big, use either larger prescaler or slower processor
    #endif
    #if (T87US < 32)
      #warning T87US too small, use either smaller prescaler or faster processor
    #endif

    TCCR0A  = (0 << COM0A1)
            | (0 << COM0A0)       // com = 00: normal mode, pin operates as usual
            | (0 << COM0B1) 
            | (0 << COM0B0)       // com = 00: normal mode, pin operates as usual
            | (0 << WGM01)        // wgm = 00: normal mode, top=0xff       
            | (0 << WGM00);       // wgm = 00: normal mode, top=0xff
          
    TCCR0B  =  (0 << FOC0A)       // force output compare
             | (0 << FOC0B)       
             | (0 << WGM02)       
             | 0;                 // stopped (T0_PRESCALER_BITS);    //   = run 

    TCNT0 = 256L - T87US;  

    // OCR0 is unused -> could be used for flags!

    
    // Init Timer1 normal Mode, tick 1us
    // Overflow at 0xFFFF (Mode0)
    
    #define TIMER1_CLKDIV 8              // possible values: 1, 8, 64, 256, 1024

    TCCR1A = (0 << COM1A1)      // compare match A
           | (0 << COM1A0) 
           | (0 << COM1B1)      // compare match B
           | (0 << COM1B0) 
           | 0                  // reserved
           | 0                  // reserved
           | (0 << WGM11)  
           | (0 << WGM10);      // Timer0 Mode 7 = Fast PWM, 10 Bit
    TCCR1B = (0 << FOC1A) 
           | (0 << FOC1B) 
           | (0 << WGM12) 
    #if   TIMER1_CLKDIV == 1
           | (0 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[2:0]=001 clkdiv 1
    #elif TIMER1_CLKDIV == 8
           | (0 << CS12) | (1 << CS11)  | (0 << CS10);       // CS[2:0]=010 clkdiv 8
    #elif TIMER1_CLKDIV == 64
           | (0 << CS12) | (1 << CS11)  | (1 << CS10);       // CS[2:0]=011 clkdiv 64 
    #elif TIMER1_CLKDIV == 256
           | (1 << CS12) | (0 << CS11)  | (0 << CS10);       // CS[2:0]=100 clkdiv 256
    #elif TIMER1_CLKDIV == 1024
           | (1 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[2:0]=101 clkdiv 1024   
    #else
     #warning: TIMER1_CLKDIV is void
    #endif


    TIMSK = (0<<TOIE1)          // Timer1 Overflow
          | (0<<OCIE1A)         // Timer1 Compare A
          | (0<<OCIE1B)         // Timer1 Compare B
          | 0                   // reserved
          | (0<<ICIE1)          // Timer1 Input Capture
          | (0<<OCIE0B)         // Timer0 Compare B
          | (0<<TOIE0)          // Timer0 Overflow
          | (0<<OCIE0A);        // Timer0 Compare A

    // Init Interrupt 0

    GIMSK = (0<<INT1)           // Int1 is msb
          | (1<<INT0)           // 1= Enable INT0 -> DCC
          | (0<<PCIE); 

    MCUCR = (0<<ISC01)
          | (1<<ISC00);         //  any edge of INT0 generates an interrupt request.
                                
    LED_DCC_POLA_P;

    semaphor_clear(C_Railcom);    // clear any railcom window
    NO_CUTOUT;    
  }

//==============================================================================
//
// Section 3 Clock Adjust
//
//==============================================================================
// a) we check timing from edge to edge, if sensible, feed to filter.
// b) we calculate a filter (rectangular coeffs)
// c) depending on output we increment or decrement OSCCAL
// runtime: approx. 15us @ 8MHz
#define PLL_LENGTH  32      // fixed - see divider
uint8_t dcc_cycle[PLL_LENGTH];
uint8_t dcc_cycle_wr_ptr;

static void init_clock_adjust(void)
  {
    uint8_t i;
    for (i=0; i<PLL_LENGTH; i++)
      {
        dcc_cycle[i] = PERIOD_1;
      }
  }


void check_clock_adjust(signed int gap)
  {
    if ((gap > PERIOD_1 - 20) && (gap < PERIOD_1 + 20)) 
      {
        dcc_cycle[dcc_cycle_wr_ptr++] = gap;
        if (dcc_cycle_wr_ptr >= PLL_LENGTH) dcc_cycle_wr_ptr = 0;
      }
    else if ((gap > PERIOD_0 - 30) && (gap < PERIOD_0 + 200)) 
      {
        // gap = gap /2;
        gap = gap >> 1;  // code is much smaller
        dcc_cycle[dcc_cycle_wr_ptr++] = gap;
        if (dcc_cycle_wr_ptr >= PLL_LENGTH) dcc_cycle_wr_ptr = 0;
      }
  }


void calc_period_1_length(void)
  {
    uint8_t i;
    unsigned int filter;
    
    filter = 0;
    for (i=0; i<PLL_LENGTH; i++)
      {
        filter += dcc_cycle[i];
      }
    filter = filter >> 3;           // /8 => 4 x period_1
    period_1_4x = filter;

    filter = filter >> 2;           // /4 => 1 x period_1
    period_1_1x = filter;
  }

//==============================================================================
//
// Section 4 Interrupts
//
//==============================================================================

ISR(TIMER1_COMPA_vect)                      // Timer1 Compare match A
  {
    if (!semaphor_query(C_Out_OFF))
      {
        DISABLE_OUT;                            // start cutout
        _delay_us(RAILCOM_DEADTIME);
        MAKE_CUTOUT;
        ENABLE_OUT;
      }
    TIMSK &= ~(1<<OCIE1A);                  // stop this Timer1 Compare A
  }

ISR(TIMER1_COMPB_vect)                      // Timer1 Compare match B
  {
    if (!semaphor_query(C_Out_OFF))
      {
        DISABLE_OUT;
        NO_CUTOUT;                              // stop cutout
        _delay_us(RAILCOM_DEADTIME);
        ENABLE_OUT;
      }
    TIMSK &= ~(1<<OCIE1B);                  // stop this Timer1 Compare B
  }

ISR(INT0_vect) 
  {
    LED_DCC_EDGE;
    dcc_edge.ui16 = TCNT1;                  // save timing reference point
    semaphor_set_isr(C_DCC_Edge);       
  }

//-------------------------------------------------------------------------------
//
// DCC Receive Routine
//
// Howto:    Edge on DCC generates an Int and dumps the timing point to last_dcc_edge.
//           The main program grab this value and check the timing difference to the
//           previous edge.
//
//                           |<-----116us----->|
//
//           DCC 1: _________XXXXXXXXX_________XXXXXXXXX_________
//                           ^-INT0
//                           |--58us--|
//                                    ^compare
//
//           DCC 0: _________XXXXXXXXXXXXXXXXXX__________________
//                           ^-INT0
//                           |-----116us----->|
//                                            ^compare
//           
// Result:   1. The received message is collected in the struct "local"
//           2. After receiving a complete message, data is copied to
//              "incoming".
//           3. The flag C_Received is set.
//
//
// Polarity Detection:
//           Look at the first 0 after preamble.
//
// Railcom Implementation
//           1. ensure correct polarity (means: invert polarity if neccessary)
//           2. reset timer2 every rising int
//           3. at endbit of message: set flag -> C_railcom
//           4. in main: check flag and correct time gap -> and send answer
//

#define T_GAP ((F_CPU / TIMER1_CLKDIV * PERIOD_1 * 5 / 4 / 1000000L)%256)  // 1.25 * PERIOD_1

// here just a repetition of the defines in dcc_receiver.h
// #define MAX_DCC_SIZE  6
// typedef struct
//   {
//     unsigned char size;               // 3 .. 6, including XOR
//     unsigned char dcc[MAX_DCC_SIZE];  // the dcc content
//   } t_message;

//!! t_message incoming;

//!! volatile t_message local;


struct
    {
        unsigned char state;                    // current state
        unsigned char bitcount;                 // current bit
        unsigned char bytecount;                // pointer to current byte
        unsigned char accubyte;                 // location for bit stuffing
        signed char dcc_time;                   // integration time for dcc (only sampling code)
                                                // we start with -7 -> all values >= indicate a zero
        unsigned char filter_data;              // bitfield for low pass data
    } dccrec;

// some states:
#define RECSTAT_WF_PREAMBLE  0
#define RECSTAT_WF_PREAMBLE1 1                  
#define RECSTAT_WF_PREAMBLE2 2
#define RECSTAT_WF_LEAD0     3
#define RECSTAT_WF_BYTE      4
#define RECSTAT_WF_TRAILER   5

#define DCCSTAT_DCC          7   
#define DCCSTAT_PHASE        6   


// ISR(INT0) loads only a register and stores this register to IO.
// this influences no status flags in SREG.
// therefore we define a naked version of the ISR with
// no compiler overhead.


void dcc_poll(void)
  {
    t_timer ocrval;
    signed int pulse;
    
    if (!semaphor_get(C_DCC_Edge)) return;              // no new Edge      
    
    pre_previous_dcc_edge.ui16 = previous_dcc_edge.ui16;
    previous_dcc_edge.ui16 = last_dcc_edge.ui16;
    last_dcc_edge.ui16 = dcc_edge.ui16;

    pulse = last_dcc_edge.i16 - previous_dcc_edge.i16;
    
    LED_DCC_EDGE_OFF;

    pulse -= T87US;   
    if (pulse < 0) DCCstate |= 1<<DCCSTAT_DCC;
    else           DCCstate &= ~(1<<DCCSTAT_DCC);
  
    #define mydcc (DCCstate & (1<<DCCSTAT_DCC))

    if (mydcc) LED_DCC_ON;
    else LED_DCC_OFF;
        
    if (DCCstate & (1<<DCCSTAT_PHASE))                  // skip second half of DCC bit
      {
        DCCstate &= ~(1<<DCCSTAT_PHASE); 
        pulse = last_dcc_edge.i16 - pre_previous_dcc_edge.i16;
        LED_PLL_CALL;
        check_clock_adjust(pulse);
        LED_PLL_CALL_OFF;
      }
    else
      {
        if (Recstate & (1<<RECSTAT_WF_PREAMBLE))        // wait for preamble
          {
            LED_DCC_END_BIT_OFF;
            if (mydcc)                                  // if polarity okay: this is normal preamble start, no BiDi
              {                                         // if inverted: could be a false detection due to start of cutout
                dccrec.bitcount++;
                Recstate = 1<<RECSTAT_WF_PREAMBLE1 ;    // check again
              }
            else
              {
                LED_BIDI_REC;
                // if we receive a single 0 at start of preamble -> could be railcom
                // railcom means: the gap must be larger than a 400us  
                //                but smaller than 580us (5 ONES)           

                if ((pulse >= 400) && (pulse <= 580))  
                  {
                    semaphor_set(C_RailcomRec);     // railcom received
                    LED_ORANGE_ON;
                  }
                else
                  {
                    semaphor_clear(C_RailcomRec);
                    LED_ORANGE_OFF;
                  }
                Recstate = 1<<RECSTAT_WF_PREAMBLE2;
              }
    	  }    
        else if (Recstate & (1<<RECSTAT_WF_PREAMBLE1))
          {
            if (mydcc)
              {
                semaphor_clear(C_RailcomRec);
                LED_ORANGE_OFF; 
                Recstate = 1<<RECSTAT_WF_PREAMBLE2 ;    // definitely no railcom
              }
            else
              {
                LED_BIDI_REC;
                // if we receive a single 0 at start of preamble -> could be railcom
                // railcom means: the gap must be larger than a 400us  
                //                but smaller than 580us (5 ONES)           

                if ((pulse >= 300) && (pulse <= 580))  
                  {
                    semaphor_set(C_RailcomRec);     // railcom received
                    LED_ORANGE_ON;
                  }
                else
                  {
                    semaphor_clear(C_RailcomRec);
                    LED_ORANGE_OFF;
                  }
                Recstate = 1<<RECSTAT_WF_PREAMBLE2;
              }
          }
        else if (Recstate & (1<<RECSTAT_WF_PREAMBLE2))            // wait for preamble (formerly the only inital state)
          {                                       
            LED_BIDI_REC_OFF;
            dccrec.bitcount++;

            if (mydcc)
              {
                //!! semaphor_get(C_Railcom);                    // we got a one - kill any existing railcom window
                //!! do not kill the railcom window, done by watch_railcom
                if (dccrec.bitcount >= 19)                  // 10 bits = 20 halfbits
                  {
                    Recstate = 1<<RECSTAT_WF_LEAD0;            
                  }
              }
            else
              {
                dccrec.bitcount=0;
              }
          }

        else if (Recstate & (1<<RECSTAT_WF_LEAD0))          // wait for leading 0
          {
            if (mydcc)
              {                                             // still 1, wait again
            
              }
            else
              {
                DCCstate |= (1<<DCCSTAT_PHASE); 
            
                dccrec.bytecount=0;
                Recstate = 1<<RECSTAT_WF_BYTE;
                dccrec.bitcount=0;
                dccrec.accubyte=0;
                LED_DCC_ST_BIT;
              }
          }
        else if (Recstate & (1<<RECSTAT_WF_BYTE))           // wait for byte
          {
            DCCstate |= (1<<DCCSTAT_PHASE); 
            LED_DCC_ST_BIT_OFF;
            unsigned char my_accubyte;
            my_accubyte = dccrec.accubyte << 1;
            if (mydcc)
              {
                my_accubyte |= 1;
              }
            dccrec.accubyte = my_accubyte;
            dccrec.bitcount++;
            if (dccrec.bitcount==8)
              {
                if (dccrec.bytecount == MAX_DCC_SIZE)       // too many bytes
                  {                                         // ignore message
                    Recstate = 1<<RECSTAT_WF_PREAMBLE;   
                  }
                else
                  {
                    //!! local.dcc[dccrec.bytecount++] = dccrec.accubyte;
                    Recstate = 1<<RECSTAT_WF_TRAILER; 
                  }
              }
          }
        else if (Recstate & (1<<RECSTAT_WF_TRAILER))        // wait for 0 (next byte) 
          {                                                 // or 1 (eof message)
            DCCstate |= (1<<DCCSTAT_PHASE);
            if (mydcc)
              {  // trailing "1" received
                Recstate = 1<<RECSTAT_WF_PREAMBLE;
                dccrec.bitcount=1;

                LED_DCC_END_BIT;
                                                            // here: enable cutout start and cutout end
                if (JUMPER_FITTED || semaphor_query(C_RailcomRec))
                // if (JUMPER_FITTED)
                  {
                    ocrval.i16 = previous_dcc_edge.i16;
                    ocrval.i16 += period_1_1x + RAILCOM_CUTOUT_START;
                    cli();
                    OCR1A = ocrval.ui16;
                    sei();
                    ocrval.i16 += period_1_4x;
                    ocrval.i16 -= (4*PERIOD_1 - RAILCOM_CUTOUT_LENGTH);
                    cli();
                    OCR1B = ocrval.ui16;
                    sei();
                    TIFR |= (1<<OCF1A) | (1<<OCF1B);     // Clear any pending int
                    TIMSK |= (1<<OCIE1A) | (1<<OCIE1B);
                  }
              }
            else
              {
                Recstate = 1<<RECSTAT_WF_BYTE;
                dccrec.bitcount=0;
                dccrec.accubyte=0;
              }
          }
        else
          {
            Recstate = 1<<RECSTAT_WF_PREAMBLE2;
          }
      }
  }


//==============================================================================
//
// Section 5 Tools
//
//==============================================================================

static inline void turn_output_off(void) 
       __attribute__((always_inline));

void turn_output_off(void)
  {
    semaphor_set(C_Out_OFF);
    DISABLE_OUT;
    MAKE_FEEDBACK;
  }

signed int timerval;
t_timer next_timer;

void watch_timer(void)
  {
    t_timer temp;
    signed int delta;
    cli();
    temp.ui16 = TCNT1;
    sei();
    delta = temp.i16 - next_timer.i16;
    if (delta >= 0)
      {
        next_timer.i16 += TICK_PERIOD;
        timerval++;
        // debug-output
        // if (timerval & 0x01) PORTD |= (1 << NU_PD6);
        // else                 PORTD &= ~(1 << NU_PD6);
      }
  }

// power-up blinking
void blink_led(void)
  {
    volatile signed int led_time;
    unsigned char i;
    
    for (i=0; i<3; i++)
      {
        LED_RED_ON; LED_ORANGE_ON;  LED_GREEN_ON;
        led_time = timerval;
        while ((timerval - led_time) < (BLINK_PERIOD / 2L / TICK_PERIOD)) watch_timer();
        LED_RED_OFF; LED_ORANGE_OFF; LED_GREEN_OFF;
        led_time = timerval;
        while ((timerval - led_time) < (BLINK_PERIOD / 2L / TICK_PERIOD)) watch_timer();
      }    
  }


unsigned char dcc_level;
unsigned char dcc_state;
signed int last_dcc_change;

// return true if a change in dcc-input is freshly detected
// the timerval for this change is stored in last_dcc_change
unsigned char check_dcc_change(void)
  {
    unsigned char retval = 0;
    // new 14.09.2009
    if (!IS_EXTON) return(retval);  // if disable with external pin - ignore DCC changes
                                    // this will shut down due to timeout. 

    if (dcc_level)
      {
        if (DCC_IN_LOW)
          {
            dcc_level = 0;
            last_dcc_change = timerval;
            if (!dcc_state) retval=1;
            dcc_state = 1;                                      // DCC is okay
            LED_GREEN_ON;
          }
      }
    else
      {
        if (DCC_IN_HIGH)
          {
            dcc_level = 1;
            last_dcc_change = timerval;
            if (!dcc_state) retval=1;
            dcc_state = 1;
            LED_GREEN_ON;            
          }
      }
    return(retval);
  }




//==============================================================================
//
// Section 6 Main
//
//==============================================================================

unsigned char fast_recover;   // Number of power-up retries

signed int last_enable;
signed int last_short;
signed int last_loaded;

unsigned char block_short;    // if true, short detection is blocked (but only for some time)
                            

int main(void)
  {
    signed int timercomp;
    init_main();
    init_clock_adjust();        // if code space is an issue - init with C
    
    turn_output_off();

/*
check_clock_adjust(PERIOD_0+2);
check_clock_adjust(PERIOD_0-2);
check_clock_adjust(PERIOD_0-4);
*/

    sei();

    blink_led();                // say hello

    while(1)
      {
        dcc_poll();
        
        calc_period_1_length();

        dcc_poll();

        watch_timer();
        if (check_dcc_change()) 
          {
            semaphor_clear(C_Out_OFF);
            ENABLE_OUT; // try to enable Output if DCC coming back
            NO_FEEDBACK;
            LED_RED_OFF;
            last_enable = timerval;
            block_short = 1;
          }

        timercomp = timerval;
        if ((timercomp - last_dcc_change) > (DCC_TIMEOUT_PERIOD / TICK_PERIOD))
          {
            dcc_state = 0;
            LED_GREEN_OFF;
            turn_output_off();
          }

        // if (IS_LOADED) { LED_ORANGE_ON; }   // old code up to V0.06
        // else { LED_ORANGE_OFF; }

        if (IS_LOADED) 
          { 
            if (IS_LED_RED_ON())
              {
                dcc_poll();                     // keep it on
              }
            else
              { 
                LED_RED_ON;                     // short pulse
                dcc_poll(); 
                LED_RED_OFF;
              }
          }
        else
          {
            dcc_poll(); 
          }

        timercomp = timerval;
            
        if (IS_SHORT)
          {
            if (block_short)
              {
                // we need to block short detection (but only for some time)
                if ((timercomp - last_enable) > (BLOCK_SHORT_TIME / TICK_PERIOD))
                    block_short = 0;
              }
            else
              {
                if (!semaphor_query(C_Out_OFF))
                  {                                 // changing from running to short
                    last_short = timercomp;         // remember time, when last short occured
                  }
                turn_output_off();
                LED_RED_ON;
              }
          }
        else
          { // not short, running
            // check, whether we are stable on - if so, set fast_recover allowed
            
            if ((timercomp - last_enable) > (2L * SLOW_RECOVER_PERIOD / TICK_PERIOD))
              {
                fast_recover = 3;
              }
            
            if (semaphor_query(C_Out_OFF))
              { // try to enable output (either fast or slow)
                if (fast_recover > 0)
                  {
                    if ((timercomp - last_short) > (FAST_RECOVER_PERIOD / TICK_PERIOD))
                      {
                        fast_recover--;
                        last_short = timercomp;   // shift reference point to actual
                        if (dcc_state)
                          {
                            // block_short = 1;          // das nicht aus fast recover heraus
                            last_enable = timercomp;
                            ENABLE_OUT;
                            semaphor_clear(C_Out_OFF);           
                          }
                      }
                  }
                else
                  {
                    if ((timercomp - last_short) > (SLOW_RECOVER_PERIOD / TICK_PERIOD))
                      {
                        last_short = timercomp;   // shift reference point to actual
                        if (dcc_state)
                          {
                            block_short = 1;
                            last_enable = timercomp;
                            ENABLE_OUT;
                            semaphor_clear(C_Out_OFF);
                          }
                      }
                  }
              }
            else
              {
                // we have the state on, and we are not short; wait a little bit until we
                // set the LED_RED to OFF
                if ((timercomp - last_enable) > (LED_RED_HOLD_TIME / TICK_PERIOD))
                  {
                    LED_RED_OFF;
                    NO_FEEDBACK;
                  }
              }
          }
      }
    return(0);  // dummy  
  }
